/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Basic model of a synchronous hardware design
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/component.hpp>
#include <hw/interface.hpp>
#include <hw/matcher.hpp>
#include <hw/message.hpp>
#include <hw/process.hpp>
#include <hw/random.hpp>
#include <hw/reaction.hpp>
#include <hw/reactions.hpp>
#include <hw/signals.hpp>
#include <hw/synchronizer.hpp>
#include <hw/time.hpp>

#include <algorithm>
#include <map>
#include <iostream>
#include <sstream>
#include <stack>
#include <vector>

#include <assert.h>
#include <stdint.h>

#ifdef CPPTESK_DEBUG_HW
    #include <stdio.h>
#endif

#if defined(WIN32)
    #include <hw/utils/win32/ucontext.h>
#elif defined(__GNUC__)
    #include <ucontext.h>
#else
    #error "unknown platform"
#endif

////////////////////////////////////////////////////////////////////////////////
// Return Types of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_PROCESS
 * Keywork marking a model process.
 */
#define CPPTESK_PROCESS void

/** @def CPPTESK_STIMULUS
 * Keywork marking a stimulus.
 */
#define CPPTESK_STIMULUS CPPTESK_PROCESS

/** @def CPPTESK_REACTION
 * Keywork marking a reaction.
 */
#define CPPTESK_REACTION CPPTESK_PROCESS

#ifdef HW_SHORT_NAMES
    #define PROCESS  CPPTESK_PROCESS
    #define STIMULUS CPPTESK_STIMULUS
    #define REACTION CPPTESK_REACTION
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Model Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MODEL
 * Declares a model of a hardware component.
 */
#define CPPTESK_MODEL(UserModel) \
    struct UserModel: public ::cpptesk::hw::Model<UserModel>

#define CPPTESK_SIGNAL_MODEL(UserModel, Inputs, Outputs) \
    struct UserModel: public ::cpptesk::hw::Model<UserModel>, public ::cpptesk::hw::IOSignals<Inputs, Outputs>

#ifdef HW_SHORT_NAMES
    #define MODEL(UserModel)                         CPPTESK_MODEL(UserModel)
    #define SIGNAL_MODEL(UserModel, Inputs, Outputs) CPPTESK_SIGNAL_MODEL(UserModel, Inputs, Outputs)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Parameters of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_PARAMETER_PROCESS process
#define CPPTESK_PARAMETER_IFACE   iface
#define CPPTESK_PARAMETER_MESSAGE message

/** @def CPPTESK_GET_PROCESS
 * Returns a reference to a process parameter.
 */
#define CPPTESK_GET_PROCESS() CPPTESK_PARAMETER_PROCESS

/** @def CPPTESK_GET_IFACE
 * Returns a reference to an interface parameter.
 */
#define CPPTESK_GET_IFACE() CPPTESK_PARAMETER_IFACE

/** @def CPPTESK_GET_MESSAGE
 * Returns a reference to a message parameter.
 */
#define CPPTESK_GET_MESSAGE() CPPTESK_PARAMETER_MESSAGE

/** @def CPPTESK_CAST_MESSAGE
 * Casts a message parameter to a given type and returns a reference.
 */
#define CPPTESK_CAST_MESSAGE(MessageType) \
    (dynamic_cast<MessageType&>(CPPTESK_PARAMETER_MESSAGE))

#define CPPTESK_CONST_CAST_MESSAGE(MessageType) \
    (dynamic_cast<const MessageType&>(CPPTESK_PARAMETER_MESSAGE))

/** @def CPPTESK_GET_PARAMETERS
 * Returns a comma-separated list of parameters.
 */
#define CPPTESK_GET_PARAMETERS() \
    CPPTESK_PARAMETER_PROCESS, CPPTESK_PARAMETER_IFACE, CPPTESK_PARAMETER_MESSAGE

#ifdef HW_SHORT_NAMES
    #define PARAMETER_PROCESS               CPPTESK_PARAMETER_PROCESS
    #define PARAMETER_IFACE                 CPPTESK_PARAMETER_IFACE
    #define PARAMETER_MESSAGE               CPPTESK_PARAMETER_MESSAGE
    #define GET_PROCESS()                   CPPTESK_GET_PROCESS()
    #define GET_IFACE()                     CPPTESK_GET_IFACE()
    #define GET_MESSAGE()                   CPPTESK_GET_MESSAGE()
    #define CAST_MESSAGE(MessageType)       CPPTESK_CAST_MESSAGE(MessageType)
    #define CONST_CAST_MESSAGE(MessageType) CPPTESK_CAST_MESSAGE(MessageType)
    #define GET_PARAMETERS()                CPPTESK_GET_PARAMETERS()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_PROCESS
 * Declares a process within a model class.
 */
#define CPPTESK_DECLARE_PROCESS(process_name) \
    virtual CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_PARAMETER_PROCESS, \
        ::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_DECLARE_NON_VIRTUAL_PROCESS
 * Declares a non-virtal process within a model class.
 */
#define CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(process_name) \
    virtual CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_PARAMETER_PROCESS, \
        ::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_DECLARE_STIMULUS
 * Declares a stimulus within a model class.
 */
#define CPPTESK_DECLARE_STIMULUS(stimulus_name) \
    virtual CPPTESK_STIMULUS stimulus_name(::cpptesk::hw::Process &CPPTESK_PARAMETER_PROCESS, \
        ::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_DECLARE_PRECONDITION
 * Declares a stimulus precondition within a model class.
 */
#define CPPTESK_DECLARE_PRECONDITION(stimulus_name) \
    virtual bool pre_##stimulus_name(::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, \
        ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_DECLARE_STIMULUS_WITH_PRE \
 * Declares a stimulus and its precondition.
 */
#define CPPTESK_DECLARE_STIMULUS_WITH_PRE(stimulus_name) \
    CPPTESK_DECLARE_PRECONDITION(stimulus_name); \
    CPPTESK_DECLARE_STIMULUS(stimulus_name)

/** @def CPPTESK_DECLARE_REACTION
 * Declares a reaction within a model class.
 */
#define CPPTESK_DECLARE_REACTION(reaction_name) \
    virtual CPPTESK_REACTION reaction_name(::cpptesk::hw::Process &CPPTESK_PARAMETER_PROCESS, \
        ::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

#ifdef HW_SHORT_NAMES
    #define DECLARE_PROCESS(process_name)             CPPTESK_DECLARE_PROCESS(process_name)
    #define DECLARE_NON_VIRTUAL_PROCESS(process_name) CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(process_name)
    #define DECLARE_STIMULUS(stimulus_name)           CPPTESK_DECLARE_STIMULUS(stimulus_name)
    #define DECLARE_PRECONDITION(stimulus_name)       CPPTESK_DECLARE_PRECONDITION(stimulus_name)
    #define DECLARE_STIMULUS_WITH_PRE(stimulus_name)  CPPTESK_DECLARE_STIMULUS_WITH_PRE(stimulus_name)
    #define DECLARE_REACTION(reaction_name)           CPPTESK_DECLARE_REACTION(reaction_name)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Special-Purpose Fields
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_IFACE
 * Declares a model interface.
 */
#define CPPTESK_DECLARE_IFACE(iface) \
    ::cpptesk::hw::Interface iface

/** @def CPPTESK_DECLARE_INPUT
 * Declares an input interface.
 */
#define CPPTESK_DECLARE_INPUT(iface) \
    ::cpptesk::hw::InputInterface iface

/** @def CPPTESK_DECLARE_OUTPUT
 * Declares an output interface.
 */
#define CPPTESK_DECLARE_OUTPUT(iface) \
    ::cpptesk::hw::OutputInterface iface

/** @def CPPTESK_DECLARE_INOUT
 * Declares an input/output interface.
 */
#define CPPTESK_DECLARE_INOUT(iface) \
    ::cpptesk::hw::InoutInterface iface

/** @def CPPTESK_DECLARE_ARBITER
 * Declares a primary reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_ARBITER(type, arbiter) \
    ::cpptesk::hw::type##ReactionMatcher arbiter

#ifdef HW_SHORT_NAMES
    #define DECLARE_IFACE(iface)           CPPTESK_DECLARE_IFACE(iface)
    #define DECLARE_INPUT(iface)           CPPTESK_DECLARE_INPUT(iface)
    #define DECLARE_OUTPUT(iface)          CPPTESK_DECLARE_OUTPUT(iface)
    #define DECLARE_INOUT(iface)           CPPTESK_DECLARE_INOUT(iface)
    #define DECLARE_ARBITER(type, arbiter) CPPTESK_DECLARE_ARBITER(type, arbiter)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Definition of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DEFINE_PROCESS
 * Defines a process.
 */
#define CPPTESK_DEFINE_PROCESS(process_name) \
    CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_PARAMETER_PROCESS, \
        ::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_DEFINE_STIMULUS
 * Defines a stimulus.
 */
#define CPPTESK_DEFINE_STIMULUS(stimulus_name) \
    CPPTESK_STIMULUS stimulus_name(::cpptesk::hw::Process &CPPTESK_PARAMETER_PROCESS, \
        ::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_DEFINE_PRECONDITION
 * Defines a stimulus precondition.
 */
#define CPPTESK_DEFINE_PRECONDITION(pre_name) \
    bool pre_name(::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, \
        ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_DEFINE_REACTION
 * Defines a reaction within a model class.
 */
#define CPPTESK_DEFINE_REACTION(reaction_name) \
    CPPTESK_REACTION reaction_name(::cpptesk::hw::Process &CPPTESK_PARAMETER_PROCESS, \
        ::cpptesk::hw::Interface &CPPTESK_PARAMETER_IFACE, ::cpptesk::hw::Message &CPPTESK_PARAMETER_MESSAGE)

#ifdef HW_SHORT_NAMES
    #define DEFINE_PROCESS(process_name)   CPPTESK_DEFINE_PROCESS(process_name)
    #define DEFINE_STIMULUS(stimulus_name) CPPTESK_DEFINE_STIMULUS(stimulus_name)
    #define DEFINE_PRECONDITION(pre_name)  CPPTESK_DEFINE_PRECONDITION(pre_name)
    #define DEFINE_REACTION(reaction_name) CPPTESK_DEFINE_REACTION(reaction_name)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Registration of Special-Purpose Components
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_ADD_COMPONENT
 * Registers a component in a model.
 */
#define CPPTESK_ADD_COMPONENT(component) \
    do { \
        component.updateName(#component); \
        registerComponent(component); \
    } while(false)

/** @def CPPTESK_ADD_IFACE
 * Registers an interface in a model.
 */
#define CPPTESK_ADD_IFACE(iface) CPPTESK_ADD_COMPONENT(iface)

/** @def CPPTESK_ADD_INPUT
 * Registers an input interface in a model.
 */
#define CPPTESK_ADD_INPUT(iface) CPPTESK_ADD_COMPONENT(iface)

/** @def CPPTESK_ADD_OUTPUT
 * Registers an output interface in a model.
 */
#define CPPTESK_ADD_OUTPUT(iface) CPPTESK_ADD_COMPONENT(iface)

/** @def CPPTESK_SET_IFACE_ADAPTER
 * Registers an interface adapter (serializer or deserializer).
 */
#define CPPTESK_SET_IFACE_ADAPTER(iface, full_method_name) \
    registerInterfaceAdapter(iface, (model_t::process_t)&full_method_name)

/** @def CPPTESK_SET_INPUT_ADAPTER
 * Registers an input interface adapter (serializer).
 */
#define CPPTESK_SET_INPUT_ADAPTER(iface, full_method_name) \
    registerInputAdapter(iface, (model_t::process_t)&full_method_name)

/** @def CPPTESK_SET_OUTPUT_ADAPTER
 * Registers an output interface adapter (deserializer).
 */
#define CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name) \
    registerOutputAdapter(iface, (model_t::process_t)&full_method_name)

/** @def CPPTESK_SET_INOUT_ADAPTER
 * Alias of CPPTESK_SET_IFACE_ADAPTER
 */
#define CPPTESK_SET_INOUT_ADAPTER(iface, full_method_name) \
    CPPTESK_SET_IFACE_ADAPTER(iface, full_method_name)

/** @def CPPTESK_SET_INNER_IFACE_ADAPTER
 * Registers an interface adapter for a inner model's interface.
 */
#define CPPTESK_SET_INNER_IFACE_ADAPTER(InnerModelClass, inner_model, inner_iface, full_method_name) \
    inner_model.registerInterfaceAdapter(inner_iface, reinterpret_cast<InnerModelClass::model_t&>(*this), \
    (InnerModelClass::model_t::process_t)&full_method_name)

/** @def CPPTESK_SET_INNER_INPUT_ADAPTER
 * Registers an input interface adapter for a inner model's interface.
 */
#define CPPTESK_SET_INNER_INPUT_ADAPTER(InnerModelClass, inner_model, inner_iface, full_method_name) \
    inner_model.registerInputAdapter(inner_iface, reinterpret_cast<InnerModelClass::model_t&>(*this), \
        (InnerModelClass::model_t::process_t)&full_method_name)

/** @def CPPTESK_SET_INNER_OUTPUT_ADAPTER
 * Registers an output interface adapter for a inner model's interface.
 */
#define CPPTESK_SET_INNER_OUTPUT_ADAPTER(InnerModelClass, inner_model, inner_iface, full_method_name) \
    inner_model.registerOutputAdapter(inner_iface, reinterpret_cast<InnerModelClass::model_t&>(*this), \
        (InnerModelClass::model_t::process_t)&full_method_name)

/** @def CPPTESK_SET_INNER_INOUT_ADAPTER
 * Alias of CPPTESK_SET_INNER_IFACE_ADAPTER
 */
#define CPPTESK_SET_INNER_INOUT_ADAPTER(inner_model, inner_iface, full_method_name) \
    CPPTESK_SET_INNER_IFACE_ADAPTER(inner_model, inner_iface, full_method_name)

/** @def CPPTESK_SET_ARBITER
 * Registers a primary arbiter for a given output interface.
 */
#define CPPTESK_SET_ARBITER(iface, arbiter) \
    iface.registerReactionMatcher(arbiter)

#ifdef HW_SHORT_NAMES
    #define ADD_COMPONENT(component)                                             CPPTESK_ADD_COMPONENT(component)
    #define ADD_IFACE(iface)                                                     CPPTESK_ADD_IFACE(iface)
    #define ADD_INPUT(iface)                                                     CPPTESK_ADD_INPUT(iface)
    #define ADD_OUTPUT(iface)                                                    CPPTESK_ADD_OUTPUT(iface)
    #define SET_IFACE_ADAPTER(iface, adapter)                                    CPPTESK_SET_IFACE_ADAPTER(iface, adapter)
    #define SET_INPUT_ADAPTER(iface, adapter)                                    CPPTESK_SET_INPUT_ADAPTER(iface, adapter)
    #define SET_OUTPUT_ADAPTER(iface, adapter)                                   CPPTESK_SET_OUTPUT_ADAPTER(iface, adapter)
    #define SET_INOUT_ADAPTER(iface, adapter)                                    CPPTESK_SET_INOUT_ADAPTER(iface, adapter)
    #define SET_INNER_IFACE_ADAPTER(inner_model, inner_iface, full_method_name)  CPPTESK_SET_INNER_IFACE_ADAPTER(inner_model, inner_iface, full_method_name)
    #define SET_INNER_INPUT_ADAPTER(inner_model, inner_iface, full_method_name)  CPPTESK_SET_INNER_INPUT_ADAPTER(inner_model, inner_iface, full_method_name)
    #define SET_INNER_OUTPUT_ADAPTER(inner_model, inner_iface, full_method_name) CPPTESK_SET_INNER_OUTPUT_ADAPTER(inner_model, inner_iface, full_method_name)
    #define SET_INNER_INOUT_ADAPTER(inner_model, inner_iface, full_method_name)  CPPTESK_SET_INNER_INOUT_ADAPTER(inner_model, inner_iface, full_method_name)
    #define SET_ARBITER(iface, arbiter)                                          CPPTESK_SET_ARBITER(iface, arbiter)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Parameters
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_SET_REACTION_TIMEOUT
 * Sets reaction timeout.
 */
#define CPPTESK_SET_REACTION_TIMEOUT(timeout) \
    setReactionTimeout(timeout)

#ifdef HW_SHORT_NAMES
    #define SET_REACTION_TIMEOUT(timeout) CPPTESK_SET_REACTION_TIMEOUT(timeout)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Control Statements
////////////////////////////////////////////////////////////////////////////////

/**
 * Process status is reassigned to EXECUTING to handle nested calls.
 */
#define __CPPTESK_SYNPOINT(type) \
    { \
        CPPTESK_PARAMETER_PROCESS.clk_count = 0; \
        CPPTESK_PARAMETER_PROCESS.syn_point = (type); \
        if(type == ::cpptesk::hw::GUARD) { CPPTESK_PARAMETER_PROCESS.timeout = false; } \
        CPPTESK_PARAMETER_PROCESS.status = ::cpptesk::hw::EXECUTING; \
    }

#define __CPPTESK_RETURN(result) \
    do { \
        CPPTESK_PARAMETER_PROCESS.status = (result); \
        return; \
    } while(false)

////////////////////////////////////////////////////////////////////////////////
// Precondition and Postcondition
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_PRE
 * Precondition assertion.
 */
#define CPPTESK_PRE(cond) \
    if(!(cond)) \
    { \
        CPPTESK_PARAMETER_PROCESS.file = __FILE__; \
        CPPTESK_PARAMETER_PROCESS.line = __LINE__; \
        __CPPTESK_RETURN(::cpptesk::hw::UNPREDICT); \
    }

/** @def CPPTESK_POST
 * Postcondition assertion.
 */
#define CPPTESK_POST(cond) \
    if(!(cond)) \
    { \
        CPPTESK_PARAMETER_PROCESS.file = __FILE__; \
        CPPTESK_PARAMETER_PROCESS.line = __LINE__; \
        __CPPTESK_RETURN(::cpptesk::hw::VIOLATION); \
    }

#ifdef HW_SHORT_NAMES
    #define PRE(cond)  CPPTESK_PRE(cond)
    #define POST(cond) CPPTESK_POST(cond)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Call Process
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_CALL_PARALLEL(method_name, iface, message) \
    start((model_t::process_t)&model_t::method_name, iface, message, START_IMMEDIATELY)

#define CPPTESK_CALL_SEQUENTIAL(method_name, iface, message) \
    start((model_t::process_t)&model_t::method_name, CPPTESK_PARAMETER_PROCESS, iface, message)

#define CPPTESK_CALL_PROCESS(call_mode, method_name, iface, message) \
    ((call_mode) == PARALLEL ? \
        CPPTESK_CALL_PARALLEL(method_name, iface, message) : \
        CPPTESK_CALL_SEQUENTIAL(method_name, iface, message))

#define CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) \
    start((model_t::process_t)&full_method_name, iface, message, mode)

#define CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message) \
    start((model_t::process_t)&full_method_name, CPPTESK_PARAMETER_PROCESS, iface, message)

#define CPPTESK_CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode) \
    ((call_mode) == PARALLEL ? \
	    /* IMPORTANT: Parallel call does not use the process parameter */ \
        CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) : \
        /* IMPORTANT: Sequential call does not use the mode parameter */ \
        CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message, mode))

#define CPPTESK_RECV_STIMULUS(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onStimulus, iface, message)

#define CPPTESK_SEND_REACTION(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
   _cpptesk_set_explanation(message, time(), time() + getReactionTimeout(), __FILE__, __LINE__))

#define CPPTESK_SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
    _cpptesk_set_explanation(message, time(), time() + getReactionTimeout(), __FILE__, __LINE__, comment))

#define CPPTESK_CALL_PROCESS_OF(model, full_method_name, iface, message) \
    (model).start(&full_method_name, iface, message)

#define CPPTESK_CALL_STIMULUS_OF(model, full_method_name, iface, message) \
    (model).start(&full_method_name, iface, message, \
    ::cpptesk::hw::Model<cpptesk::hw::EMPTY_MODEL>::START_REQUEST_IFACE)

/** @def CPPTESK_CALL_LOOP
 * Launches an infinite loop.
 */
#define CPPTESK_CALL_LOOP(method_name) \
    CPPTESK_CALL_PARALLEL(method_name, CPPTESK_DEFAULT_IFACE, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE)

/** @def CPPTESK_CALL_LOOP_EX
 * Launches an infinite loop (usually, a listener of an output interface).
 */
#define CPPTESK_CALL_LOOP_EX(full_method_name, iface, mode) \
    CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE, mode)

#ifdef HW_SHORT_NAMES
    #define CALL_PARALLEL(method_name, iface, message)                          CPPTESK_CALL_PARALLEL(method_name, iface, message)
    #define CALL_SEQUENTIAL(method_name, iface, message)                        CPPTESK_CALL_SEQUENTIAL(method_name, iface, message)
    #define CALL_PROCESS(call_mode, method_name, iface, message)                CPPTESK_CALL_PROCESS(call_mode, method_name, iface, message)
    #define CALL_PARALLEL_EX(full_method_name, iface, message, mode)            CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode)
    #define CALL_SEQUENTIAL_EX(full_method_name, iface, message)                CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message)
    #define CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode)  CPPTESK_CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode)
    #define RECV_STIMULUS(call_mode, iface, message)                            CPPTESK_RECV_STIMULUS(call_mode, iface, message)
    #define SEND_REACTION(call_mode, iface, message)                            CPPTESK_SEND_REACTION(call_mode, iface, message)
    #define SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment)      CPPTESK_SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment)
    #define CALL_PROCESS_OF(model, full_method_name, iface, message)            CPPTESK_CALL_PROCESS_OF(model, full_method_name, iface, message)
    #define CALL_STIMULUS_OF(model, full_method_name, iface, message)           CPPTESK_CALL_STIMULUS_OF(model, full_method_name, iface, message)
    #define CALL_LOOP(method_name)                                              CPPTESK_CALL_LOOP(method_name)
    #define CALL_LOOP_EX(full_method_name, iface, mode)                         CPPTESK_CALL_LOOP_EX(full_method_name, iface, mode)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Start and Stop
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_START
 * Starts a model process.
 */
#define CPPTESK_START() \
    /* Only one CPPTESK_START() is allowed in a specification function */ \
    $CPPTESK_START: while(false) { goto $CPPTESK_START; }

/** @def CPPTESK_START_PROCESS
 * Starts a process.
 */
#define CPPTESK_START_PROCESS() \
    CPPTESK_START(); \

/** @def CPPTESK_START_STIMULUS
 * Starts a stimulus and launches the input interface adapter.
 */
#define CPPTESK_START_STIMULUS(call_mode) \
    CPPTESK_START(); \
    CPPTESK_RECV_STIMULUS(call_mode, CPPTESK_PARAMETER_IFACE, CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_START_REACTION
 * Starts a reaction and launches the output interface adapter.
 */
#define CPPTESK_START_REACTION(call_mode) \
    CPPTESK_START(); \
    CPPTESK_SEND_REACTION(call_mode, CPPTESK_PARAMETER_IFACE, CPPTESK_PARAMETER_MESSAGE)

/** @def CPPTESK_STOP
 * Stops a model process.
 */
#define CPPTESK_STOP() \
    do { \
        __CPPTESK_RETURN(::cpptesk::hw::COMPLETED); \
    } while(false)

/** @def CPPTESK_STOP_PROCESS
 * Stops a process.
 */
#define CPPTESK_STOP_PROCESS()  CPPTESK_STOP()

/** @def CPPTESK_STOP_STIMULUS
 * Stops a stimulus.
 */
#define CPPTESK_STOP_STIMULUS() CPPTESK_STOP()

/** @def CPPTESK_STOP_REACTION
 * Stops a reaction.
 */
#define CPPTESK_STOP_REACTION() CPPTESK_STOP()

#ifdef HW_SHORT_NAMES
    #define START()                   CPPTESK_START()
    #define START_PROCESS()           CPPTESK_START_PROCESS()
    #define START_STIMULUS(call_mode) CPPTESK_START_STIMULUS(call_mode)
    #define START_REACTION(call_mode) CPPTESK_START_REACTION(call_mode)
    #define STOP()                    CPPTESK_STOP()
    #define STOP_PROCESS()            CPPTESK_STOP_PROCESS()
    #define STOP_STIMULUS()           CPPTESK_STOP_STIMULUS()
    #define STOP_REACTION()           CPPTESK_STOP_REACTION()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Cycle, Delay, and Guard
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DELAY
 * Emulates an internal clock cycle of a unit.
 */
#define CPPTESK_CYCLE_EX(unit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE); \
        (unit).suspend(&CPPTESK_PARAMETER_PROCESS); \
    } while(false)

/** @def CPPTESK_DELAY
 * Emulates a clock cycle.
 */
#define CPPTESK_CYCLE() \
    CPPTESK_CYCLE_EX(*this)

/** @def CPPTESK_DELAY
 * Emulates several clock cycles.
 */
#define CPPTESK_DELAY(delay) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE); \
        while(CPPTESK_PARAMETER_PROCESS.clk_count < (::cpptesk::hw::cycle_t)(delay)) \
            { suspend(&CPPTESK_PARAMETER_PROCESS); } \
    } while(false)

/** @def CPPTESK_GUARD
 * Waits until a condition becomes true.
 */
#define CPPTESK_GUARD(condition) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::GUARD); \
        while(!(condition)) \
            { suspend(&CPPTESK_PARAMETER_PROCESS); } \
        CPPTESK_PARAMETER_PROCESS.syn_point = ::cpptesk::hw::CYCLE; \
    } while(false)

/** @def CPPTESK_GUARD_TIMEOUT
 * Waits until a condition becomes true or time is over.
 */
#define CPPTESK_GUARD_TIMEOUT(condition, limit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::GUARD); \
        while(!(condition)) \
        { \
            if(CPPTESK_PARAMETER_PROCESS.clk_count >= (::cpptesk::hw::cycle_t)(limit)) \
            { \
                CPPTESK_PARAMETER_PROCESS.timeout =   true; \
                CPPTESK_PARAMETER_PROCESS.file    = __FILE__; \
                CPPTESK_PARAMETER_PROCESS.line    = __LINE__; \
                onTimeoutViolation(CPPTESK_PARAMETER_PROCESS); \
                break; \
            } \
            suspend(&CPPTESK_PARAMETER_PROCESS); \
        } \
        CPPTESK_PARAMETER_PROCESS.syn_point = ::cpptesk::hw::CYCLE; \
    } while(false)

/** @def CPPTESK_WAIT
 * Alias of CPPTESK_GUARD
 */
#define CPPTESK_WAIT(condition) \
    CPPTESK_GUARD(condition)

/** @def CPPTESK_WAIT_TIMEOUT
 * Alias of CPPTESK_GUARD_TIMEOUT
 */
#define CPPTESK_WAIT_TIMEOUT(condition, limit) \
    CPPTESK_GUARD_TIMEOUT(condition, limit)

#ifdef HW_SHORT_NAMES
    #define CYCLE_EX(unit)                  CPPTESK_CYCLE_EX(unit)
    #define CYCLE()                         CPPTESK_CYCLE()
    #define DELAY(delay)                    CPPTESK_DELAY(delay)
    #define GUARD(condition)                CPPTESK_GUARD(condition)
    #define GUARD_TIMEOUT(condition, limit) CPPTESK_GUARD_TIMEOUT(condition, limit)
    #define WAIT(condition)                 CPPTESK_WAIT(condition)
    #define WAIT_TIMEOUT(condition, limit)  CPPTESK_WAIT_TIMEOUT(condition, limit)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Waiting Process
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_WAIT_PROCESS_STARTED(pid) \
    CPPTESK_WAIT(isFinished(pid) || isStarted(pid))

#define CPPTESK_WAIT_PROCESS_STARTED_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isStarted(pid), limit)

#define CPPTESK_WAIT_PROCESS_WAITING(pid) \
    CPPTESK_WAIT(isFinished(pid) || isWaiting(pid))

#define CPPTESK_WAIT_PROCESS_WAITING_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isWaiting(pid), limit)

#define CPPTESK_WAIT_PROCESS_ACTIVE(pid) \
    CPPTESK_WAIT(isFinished(pid) || isActive(pid))

#define CPPTESK_WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isActive(pid, limit))

#define CPPTESK_WAIT_PROCESS_FINISHED(pid) \
    CPPTESK_WAIT(isFinished(pid))

#define CPPTESK_WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid), limit)

#ifdef HW_SHORT_NAMES
    #define WAIT_PROCESS_STARTED(pid)                 CPPTESK_WAIT_PROCESS_STARTED(pid)
    #define WAIT_PROCESS_STARTED_TIMEOUT(pid, limit)  CPPTESK_WAIT_PROCESS_STARTED_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_WAITING(pid)                 CPPTESK_WAIT_PROCESS_WAITING(pid)
    #define WAIT_PROCESS_WAITING_TIMEOUT(pid, limit)  CPPTESK_WAIT_PROCESS_WAITING_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_ACTIVE(pid)                  CPPTESK_WAIT_PROCESS_ACTIVE(pid)
    #define WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit)   CPPTESK_WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_FINISHED(pid)                CPPTESK_WAIT_PROCESS_FINISHED(pid)
    #define WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit) CPPTESK_WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Process Attributes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GET_PRIORITY
 * Gets a process priority.
 */
#define CPPTESK_GET_PRIORITY() \
    ((unsigned)(CPPTESK_PARAMETER_PROCESS.getPriority()))

/** @def CPPTESK_SET_PRIORITY
 * Sets a process priority.
 */
#define CPPTESK_SET_PRIORITY(priority) \
    do { \
        CPPTESK_PARAMETER_PROCESS.setPriority((::cpptesk::hw::priority_t)priority); \
    } while(false)

/** @def CPPTESK_GET_INPUT_IFACE
 * Gets a process input interface.
 */
#define CPPTESK_GET_INPUT_IFACE() \
    (CPPTESK_PARAMETER_PROCESS.input_iface)

/** @def CPPTESK_SET_INPUT_IFACE
 * Sets a process input interface.
 */
#define CPPTESK_SET_INPUT_IFACE(iface) \
    do { \
        CPPTESK_PARAMETER_PROCESS.input_iface = &iface; \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define GET_PRIORITY()         CPPTESK_GET_PRIORITY()
    #define SET_PRIORITY(priority) CPPTESK_SET_PRIORITY(priority)
    #define GET_INPUT_IFACE()      CPPTESK_GET_INPUT_IFACE()
    #define SET_INPUT_IFACE(iface) CPPTESK_SET_INPUT_IFACE(iface)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Reaction Arbitration
///////////////////////////////////////////////////////////////////////////////

/**
 * @def CPPTESK_DETECT_REACTION
 * Detects a reaction by using a single-cycle condition but do not take it.
 * It is used when a detector is multi-cycle.
 */
#define CPPTESK_DETECT_REACTION(cond) \
    do { \
        CPPTESK_WAIT_TIMEOUT(CPPTESK_PARAMETER_IFACE.isReady() || CPPTESK_PARAMETER_IFACE.getOwner() == &CPPTESK_PARAMETER_PROCESS, getReactionTimeout()); \
        CPPTESK_PARAMETER_IFACE.appendRequest(CPPTESK_PARAMETER_PROCESS, CPPTESK_PARAMETER_MESSAGE); \
        CPPTESK_WAIT_TIMEOUT((cond) && CPPTESK_PARAMETER_IFACE.chooseRequest(CPPTESK_PARAMETER_PROCESS, CPPTESK_PARAMETER_MESSAGE), getReactionTimeout()); \
    } while(false)

/**
 * @def CPPTESK_TAKE_REACTION
 * Takes a reaction (says to the interface that the reaction is taken).
 */
#define CPPTESK_TAKE_REACTION() \
    do { \
        /* Error if a reaction has been taken by someone else */ \
        if(!CPPTESK_PARAMETER_IFACE.takeReaction()) { \
            CPPTESK_ASSERTION(false, "Reaction has already been taken " \
                "(use CPPTESK_SET_ARBITER() to set a reaction arbiter for the interface)"); \
        } \
    } while(false)

/** @def CPPTESK_WAIT_REACTION
 * Detects a reaction by using a single-cycle condition and takes it.
 */
#define CPPTESK_WAIT_REACTION(cond) \
    do { \
        CPPTESK_DETECT_REACTION(cond); \
        CPPTESK_TAKE_REACTION(); \
    } while(false)

/** @def CPPTESK_NEXT_REACTION
 * Finishes processing of a reaction and switches to the next one.
 */
#define CPPTESK_NEXT_REACTION() \
    do { \
        ::cpptesk::hw::priority_t __priority = (::cpptesk::hw::priority_t)(CPPTESK_GET_PRIORITY()); \
        CPPTESK_SET_PRIORITY(::cpptesk::hw::HIGHEST_PRIORITY); \
        CPPTESK_CYCLE(); \
        CPPTESK_SET_PRIORITY(__priority); \
        CPPTESK_PARAMETER_IFACE.removeRequest(CPPTESK_PARAMETER_PROCESS, CPPTESK_PARAMETER_MESSAGE); \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define DETECT_REACTION(cond) CPPTESK_DETECT_REACTION(cond)
    #define TAKE_REACTION()       CPPTESK_TAKE_REACTION()
    #define WAIT_REACTION(cond)   CPPTESK_WAIT_REACTION(cond)
    #define NEXT_REACTION()       CPPTESK_NEXT_REACTION()
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Debug
///////////////////////////////////////////////////////////////////////////////

#define CPPTESK_DEBUG_PRINT(level, message) \
    do { \
        if(_debug_level >= (level)) \
        { \
            std::cout << "------------------------------------------------------------" << std::endl; \
            std::cout << message << std::endl; \
        } \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define DEBUG_PRINT(message) CPPTESK_DEBUG_PRINT(message)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// The Basic Model Class
///////////////////////////////////////////////////////////////////////////////

namespace cpptesk {
namespace hw {

template<class UserModel /* Subclass */> class Model: public Synchronizer
{
public:
    typedef UserModel model_t;

    const static int DEFAULT_REACTION_TIMEOUT = 10;

    typedef void (UserModel::*process_t)(Process &process, Interface &iface, Message &message);

    typedef process_t stimulus_t;
    typedef process_t reaction_t;

    enum call_mode_t { PARALLEL, SEQUENTIAL };

    enum start_mode_t
    {
        /// Calling start() adds a new process into the pipeline; the process will be executed when cycle() is called.
        START_NORMAL        = 0x0000 | NULL_PRIORITY,
        /// Calling start() adds a new process into the pipeline and executes the first stage of the process within the current cycle.
        START_IMMEDIATELY   = 0x0100 | NULL_PRIORITY,
        /// Request interface when starting a process.
        START_REQUEST_IFACE = 0x0200 | NULL_PRIORITY,
    };

    Model();
    virtual ~Model();

    void registerComponent(const Component &component);

    void registerInterfaceAdapter(Interface &iface, process_t adapter);

    inline void registerInputAdapter(Interface &input, process_t serializer)
        { registerInterfaceAdapter(input, serializer); }

    inline void registerOutputAdapter(Interface &output, process_t deserializer)
        { registerInterfaceAdapter(output, deserializer); }

    void registerInterfaceAdapter(Interface &iface, model_t &model, process_t adapter);

	inline void registerInputAdapter(Interface &input, model_t &model, process_t adapter)
	    { registerInterfaceAdapter(input, model, adapter); }

	inline void registerOutputAdapter(Interface &output, model_t &model, process_t adapter)
		{ registerInterfaceAdapter(output, model, adapter); }

    /**
     * Starts a new process for a given stimulus.
     */
    pid_t start(stimulus_t stimulus, Interface &iface, Message &message, int mode = START_NORMAL);

    /**
     * Starts a new process for a given stimulus with the empty message.
     */
    pid_t start(stimulus_t stimulus, Interface &iface, int mode = START_NORMAL);

    /**
     * Processes a stimulus within a given process. Calling start(&UserModel::stimulus, process, message)
     * is fully equivalent to userModel.stimulus(process, iface, message).
     */
    pid_t start(stimulus_t stimulus, Process &process, Interface &iface, Message &message = CPPTESK_DEFAULT_MESSAGE);

    /**
     * Stops a process.
     */
    void stop(pid_t pid);

    bool isStarted(pid_t pid)   const;
    bool isWaiting(pid_t pid)   const;
    bool isActive(pid_t pid)    const;
    bool isFinished(pid_t pid)  const;

    cycle_t time() const;
    size_t size() const;

    void cycle();
    void reset();

    void delay(cycle_t n);

    /**
     * Register a reaction.
     */
    void registerReaction(const Reaction &reaction);
    void registerReaction(error_type_t error, uint32_t type, const Interface &iface, const std::string &info);
    void registerReaction(error_type_t error, const Interface &iface, const std::string &info);
    void clearReactions();

    /**
    * Returns the first error (if there is any) having occured starting from the previous call of verdict().
    * Do not clear the set of reactions.
    */
    std::string error(error_mode_t mode = LAST_REACTION) const;

    /**
     * Returns a verdict for a set of reactions having occured starting from the previous call. 
     * Clears the set of reactions to be analyzed.
     */
    bool verdict();

    /**
     * Returns a list of reactions.
     * @depricated
     */
    const std::vector<Reaction>& getReactions() const;

////////////////////////////////////////////////////////////////////////////////
// Trace functions
////////////////////////////////////////////////////////////////////////////////

    void traceReactionStatistics();

protected:
////////////////////////////////////////////////////////////////////////////////
// Working with input and output interfaces
////////////////////////////////////////////////////////////////////////////////

    /// Pseudo interface for reactions that are not assosiated with existing interfaces.
    CPPTESK_DECLARE_INOUT(CPPTESK_DEFAULT_IFACE);

    /**
     * Processes a stimulus: usually, serializes a message and applies it to an input interface.
     * To be overloaded in a mediator.
     */
    CPPTESK_DECLARE_PROCESS(onStimulus);

    /**
     * Processes a reaction: usually, deserializes a message from an output interface and calls a comparator.
     * To be overloaded in a mediator.
     */
    CPPTESK_DECLARE_PROCESS(onReaction);

    void setReactionTimeout(unsigned timeout);

    unsigned getReactionTimeout() const;

////////////////////////////////////////////////////////////////////////////////
// Callbacks
////////////////////////////////////////////////////////////////////////////////
    virtual void onPreconditionViolation(Process &process);

    virtual void onPostconditionViolation(Process &process);

    virtual void onTimeoutViolation(Process &process);

    virtual void onNormalCompletion(Process &process);

    virtual void onEveryCycle(Process &process);

    /**
     * Checks data integrity of a model. The callback is overloaded by a model.
     */
    virtual bool invariant();

////////////////////////////////////////////////////////////////////////////////
// Synchronization methods
////////////////////////////////////////////////////////////////////////////////

public:

    /**
     * Initializes simulation.
     */
    virtual void initialize();

    /**
     * Finalizes simulation.
     */
    virtual void finalize();

    /**
     * Inputs synchronizer: sets input signals of a design.
     */
    virtual void setInputs();

    /**
     * Outputs synchronizer: gets output signals of a design.
     */
    virtual void getOutputs();

    /**
     * Time synchronizer: synchronizes execution threads of a model and simulator.
     */
    virtual void simulate();

    /**
     * Registers synchronizer.
     */
    void setSynchronizer(Synchronizer *synchronizer);

private:
////////////////////////////////////////////////////////////////////////////////
// Auxiliary methods
////////////////////////////////////////////////////////////////////////////////
    static void launch(UserModel *model, stimulus_t stimulus, Process *process, Interface *iface, Message *message);

// IMPORTANT: Should be public (CYCLE_EX)
public:
    Process* fork(Interface &iface, Message &message);
    void resume(Process *process);
    void suspend(Process *process);

protected:
////////////////////////////////////////////////////////////////////////////////
// Data
////////////////////////////////////////////////////////////////////////////////
    typedef std::vector<Component*> components_list_t;
    components_list_t _components;

    class functor_t
    {
    public:
        functor_t(model_t &model, process_t process): _model(model), _process(process) {}
        CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(call) { (_model.*_process)(CPPTESK_GET_PARAMETERS()); }
    private:
        model_t &_model;
        process_t _process;
    };

    typedef std::map<Interface*, functor_t*> adapters_map_t;
    adapters_map_t _adapters;

    cycle_t _reaction_timeout;

    ucontext_t _main_context;

    pid_t _new_pid;

    cycle_t _clock;

    typedef std::map<pid_t, Process*> process_table_t;
    process_table_t _process_table;

    Reactions _reactions;

    typedef std::stack<ucontext_t*> process_stack_t;
    process_stack_t _process_stack;

////////////////////////////////////////////////////////////////////////////////
// Debug
////////////////////////////////////////////////////////////////////////////////
protected:

    enum debug_level_t
    {
        /// Debug all reactions.
        DEBUG_ALL   = 2,
        DEBUG_INFO  = 2,
        /// Debug error reactions only.
        DEBUG_ERROR = 1,
        /// Turn off debug printing.
        DEBUG_NONE  = 0,
    };

    void debug(debug_level_t debug_level = DEBUG_NONE)
        { _debug_level = debug_level; }

    debug_level_t _debug_level;

////////////////////////////////////////////////////////////////////////////////
// Simulator Synchronizer
////////////////////////////////////////////////////////////////////////////////
private:

    Synchronizer *_synchronizer;
};

template<class UserModel>
Model<UserModel>::Model():
    CPPTESK_DEFAULT_IFACE("<default>"),
    _reaction_timeout(DEFAULT_REACTION_TIMEOUT),
    _new_pid(0),
    _clock(0),
    _synchronizer(0)
{
    _process_stack.push(&_main_context);
}

template<class UserModel>
Model<UserModel>::~Model()
{
    // Reset the pipeline
    reset();

    // Delete the interface adapters
    for(typename adapters_map_t::iterator i = _adapters.begin(); i != _adapters.end(); i++)
        { delete i->second; }
    _adapters.clear();

    // Delete the synchronizer
    if(_synchronizer)
        { delete _synchronizer; }
}

template<class UserModel>
void Model<UserModel>::registerComponent(const Component &component)
{
#ifdef CPPTESK_DEBUG_HW
    printf("Model::registerComponent\n");
#endif
    _components.push_back(const_cast<Component*>(&component));
}

template<class UserModel>
void Model<UserModel>::registerInterfaceAdapter(Interface &iface, process_t adapter)
{
	registerInterfaceAdapter(iface, dynamic_cast<model_t&>(*this), adapter);
}

template<class UserModel>
void Model<UserModel>::registerInterfaceAdapter(Interface &iface, model_t &model, process_t adapter)
{
    typename adapters_map_t::iterator i = _adapters.find(&iface);

    // Deallocate the previously registered adapter
    if(i != _adapters.end() && !(i->second))
        { delete i->second; }	

    // Register the new one
    _adapters[&iface] = new functor_t(model, adapter);
}

template<class UserModel>
void Model<UserModel>::launch(UserModel *model, stimulus_t stimulus, Process *process, Interface *iface, Message *message)
{
#ifdef CPPTESK_DEBUG_HW
    printf("Model::launch: model    = %p\n", model);
    printf("Model::launch: stimulus = %p\n", stimulus);
    printf("Model::launch: process  = %p\n", process);
    printf("Model::launch: iface    = %p\n", iface);
    printf("Model::launch: message  = %p\n", message);
    fflush(stdout);
#endif

    CPPTESK_ASSERTION(process != NULL, "Parameter process is null");
    CPPTESK_ASSERTION(iface   != NULL, "Parameter iface is null");
    CPPTESK_ASSERTION(message != NULL, "Parameter message is null");

    (model->*stimulus)(*process, *iface, *message);
}

template<class UserModel>
pid_t Model<UserModel>::start(stimulus_t stimulus, Interface &iface, Message &message, int mode)
{
    Process *process = fork(iface, message);
    CPPTESK_ASSERTION(process != NULL, "Process has not forked");

	priority_t priority = (priority_t)(mode & PRIORITY_MASK); 
    process->setPriority(priority == NULL_PRIORITY ? NORMAL_PRIORITY : priority);

    mode &= ~PRIORITY_MASK;

    if((mode & START_REQUEST_IFACE) != 0 && iface.isInput())
        { iface.request(*process, *(process->input_message)); }

    Interface *input_iface = process->input_iface;
    Message *input_message = process->input_message;

    process->ucp->uc_link = &_main_context;
    makecontext(process->ucp, (void(*)(void))Model<UserModel>::launch, 6, this, stimulus, process, input_iface, input_message);

    // Execute the first stage of the process within the current cycle if START_IMMEDIATELY is set
    if((mode & START_IMMEDIATELY) != 0)
        { resume(process); }

    return process->pid;
}

template<class UserModel>
pid_t Model<UserModel>::start(stimulus_t stimulus, Interface &iface, int mode)
{
    return start(stimulus, iface, CPPTESK_DEFAULT_MESSAGE, mode);
}

template<class UserModel>
pid_t Model<UserModel>::start(stimulus_t stimulus, Process &process, Interface &iface, Message &message)
{
    ((UserModel*)this->*stimulus)(process, iface, message);
    return process.pid;
}

template<class UserModel>
void Model<UserModel>::stop(pid_t pid)
{
    process_table_t::iterator i = _process_table.find(pid);

    if(i != _process_table.end())
    {
        delete i->second;
        _process_table.erase(i);
    }
}

template<class UserModel>
bool Model<UserModel>::isStarted(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isStarted();
}

template<class UserModel>
bool Model<UserModel>::isWaiting(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isWaiting();
}

template<class UserModel>
bool Model<UserModel>::isActive(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isActive();
}

template<class UserModel>
bool Model<UserModel>::isFinished(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);

    return i == _process_table.end();
}

template<class UserModel>
cycle_t Model<UserModel>::time() const
{
    return _clock;
}

template<class UserModel>
size_t Model<UserModel>::size() const
{
    return _process_table.size();
}

template<class UserModel>
void Model<UserModel>::cycle()
{
    std::vector<pid_t> to_kill;
    std::vector<Process*> processes;

    // Check a model's invariant
    CPPTESK_ASSERTION(invariant(), "Model invariant violation");

    // Process components
    for(components_list_t::iterator i = _components.begin(); i != _components.end(); i++)
        { (*i)->cycle(); }

    // List of processes
    for(process_table_t::iterator i = _process_table.begin(); i != _process_table.end(); i++)
        { processes.push_back(i->second); }

    // Sort processes in order of decreasing priorities
    stable_sort(processes.begin(), processes.end(), greaterProcessPriority);

    for(std::vector<Process*>::iterator i = processes.begin(); i != processes.end(); i++)
    {
        Process *process = *i;

        // The process status after starting is usually EXECUTING, but there are some exceptions
        if(process->status == EXECUTING)
            { resume(process); }

        if(process->status != EXECUTING)
        {
            CPPTESK_ASSERTION(_process_stack.top() == process->ucp, "Execution flow inconsistency");
            _process_stack.pop();

            to_kill.push_back(process->pid);
        }
    }

    for(std::vector<pid_t>::iterator j = to_kill.begin(); j != to_kill.end(); j++)
        { stop(*j); }

    // Set input signals
    setInputs();

    // Simulate a cycle 
    simulate();
    _clock++;

    // Get output signals
    getOutputs();
}

template<class UserModel>
void Model<UserModel>::reset()
{
    for(process_table_t::iterator i = _process_table.begin(); i != _process_table.end(); i++)
        { delete i->second; }

    _process_table.clear();
}

template<class UserModel>
void Model<UserModel>::delay(cycle_t n)
{
    for(cycle_t i = 0; i < n; i++)
        { cycle(); }
}

template<class UserModel>
const std::vector<Reaction>& Model<UserModel>::getReactions() const
{
    return _reactions.getReactions();
}

////////////////////////////////////////////////////////////////////////////////
// Trace functions
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
void Model<UserModel>::traceReactionStatistics()
{
    unsigned max_width = 0;
    unsigned all_count = 0;
    std::ostream &out = std::cout;

    for(unsigned error = 0; error < __ERROR_TYPE_NUMBER; error++)
    {
        const std::string &error_string = getErrorString((error_type_t)error);
        if(error_string.length() > max_width) { max_width = error_string.length(); }
    }   

    for(unsigned error = 0; error < __ERROR_TYPE_NUMBER; error++)
    {
        out.width(max_width);
        out << std::left << getErrorString((error_type_t)error) << ": ";
        out << std::dec << _reactions.count((error_type_t)error) << std::endl;
        all_count += _reactions.count((error_type_t)error);
    }

    out.width(max_width);
    out << std::left << "[TOTAL]" << ": ";
    out << std::dec << all_count << std::endl;
}

////////////////////////////////////////////////////////////////////////////////
// Working with input and output interfaces
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
CPPTESK_DEFINE_PROCESS(Model<UserModel>::onStimulus)
{
    // If a model does not have interface adapters, then do nothing
    if(_adapters.empty())
        { return; }

    // Otherwise, try to find an appropriate one
    typename adapters_map_t::iterator i = _adapters.find(&iface);

    // Ensure that an interface adapter has been found
    CPPTESK_ASSERTION(i != _adapters.end(), "Adapter has not been found");
    CPPTESK_ASSERTION(i->second != 0, "Adapter is null");

    // Call the serializer having been found
    (i->second->call)(process, iface, message /* input */);
}

template<class UserModel>
CPPTESK_DEFINE_PROCESS(Model<UserModel>::onReaction)
{
    // If a model does not have interface adapters,
    // then register a normal (correct) reaction
    if(_adapters.empty())
        { registerReaction(NORMAL_REACTION, iface, ""); return; }

#ifndef CPPTESK_CLONE_INPUT_MESSAGE
    // The default handler requies clonning of input messages
    CPPTESK_ASSERTION(false, "Default reaction handler does not support non-cloning messages");
#endif

    // Otherwise (there are interface adapters), try to find an appropriate one
    typename adapters_map_t::iterator i = _adapters.find(&iface);

    // Ensure that an interface adapter has been found
    CPPTESK_ASSERTION(i != _adapters.end(), "Adapter has not been found");
    CPPTESK_ASSERTION(i->second != 0, "Adapter is null");

    SET_PRIORITY(message.getPriority());

    if((message.getInterlock() & LOCK_HOLD) != 0)
        { WAIT_PROCESS_FINISHED(message.getPID()); }
    if((message.getInterlock() & LOCK_DELAY) != 0)
        { DELAY(message.getDelay()); }

    Message &spec_msg =  (message);
    Message &impl_msg = *(message.clone());

    // Call the deserializer having been found
    (i->second->call)(process, iface, impl_msg /* output */);

    std::stringstream buffer;

    error_type_t error = NORMAL_REACTION;

    if(process.timeout)
    {
        error = MISSING_REACTION;
    }
    else
    {
        std::string info = spec_msg.compare(impl_msg);
    
        if(!info.empty())
        {
            error = INCORRECT_REACTION;
            buffer << info << " ";
        }
    }

    if(error != NORMAL_REACTION)
        { buffer << spec_msg.getExplanation(); }

    // Register reaction
    registerReaction(error, iface, buffer.str());

    // Delete the clone
    delete &impl_msg;
}

////////////////////////////////////////////////////////////////////////////////
// Reactions
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
void Model<UserModel>::registerReaction(const Reaction &reaction)
{
    CPPTESK_DEBUG_PRINT((reaction.isError() ? DEBUG_ERROR : DEBUG_INFO), reaction);

    _reactions.registerReaction(reaction);
}

template<class UserModel>
void Model<UserModel>::registerReaction(error_type_t error, uint32_t type, const Interface &iface, const std::string &info)
{
    Reaction reaction(error, type, time(), iface.getName(), info);
    registerReaction(reaction);
}

template<class UserModel>
void Model<UserModel>::registerReaction(error_type_t error, const Interface &iface, const std::string &info)
{
    registerReaction(error, -1, iface, info);
}

template <class UserModel>
void Model<UserModel>::clearReactions()
{
    _reactions.clearReactions();
}

template <class UserModel>
std::string Model<UserModel>::error(error_mode_t mode) const
{
    return _reactions.error(mode);
}

/**
 * Returns a verdict for a set of reactions having been occured starting from the previous call. 
 */
template <class UserModel>
bool Model<UserModel>::verdict()
{
    return _reactions.verdict();
}

template<class UserModel>
void Model<UserModel>::setReactionTimeout(unsigned timeout)
{
    _reaction_timeout = timeout;
}

template<class UserModel>
unsigned Model<UserModel>::getReactionTimeout() const
{
    return _reaction_timeout;
}

////////////////////////////////////////////////////////////////////////////////
// Callbacks
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
void Model<UserModel>::onPreconditionViolation(Process &process)
{
#ifdef CPPTESK_DEBUG_HW
    printf("Model::onPreconditionViolation: pid=%d\n", process.pid);
#endif

    std::stringstream buffer;
    buffer << "(" << process.file << ", " << process.line << ")";
    registerReaction(ASSERTION_VIOLATION, CPPTESK_DEFAULT_IFACE, buffer.str());
}

template<class UserModel>
void Model<UserModel>::onPostconditionViolation(Process &process)
{
#ifdef CPPTESK_DEBUG_HW
    printf("Model::onPostconditionViolation: pid=%d\n", process.pid);
#endif

    std::stringstream buffer;
    buffer << "(" << process.file << ", " << process.line << ")";
    registerReaction(ASSERTION_VIOLATION, CPPTESK_DEFAULT_IFACE, buffer.str());
}

template<class UserModel>
void Model<UserModel>::onTimeoutViolation(Process &process)
{
#ifdef CPPTESK_DEBUG_HW
    printf("Model::onTimeoutViolation: pid=%d\n", process.pid);
#endif
    // Do nothing
}

template<class UserModel>
void Model<UserModel>::onNormalCompletion(Process &process)
{
#ifdef CPPTESK_DEBUG_HW
    printf("Model::onNormalCompletion: pid=%d\n", process.pid);
#endif
    // Do nothing
}

template<class UserModel>
void Model<UserModel>::onEveryCycle(Process &process)
{
#ifdef CPPTESK_DEBUG_HW
    printf("Model::onEveryCycle: pid=%d\n", process.pid);
#endif
    // Do nothing
}

template<class UserModel>
bool Model<UserModel>::invariant()
{
#ifdef CPPTESK_DEBUG_HW
    printf("Model::invariant\n");
#endif
    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Synchronization methods
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
void Model<UserModel>::initialize()
{
    if(_synchronizer) { _synchronizer->initialize(); }
}

template<class UserModel>
void Model<UserModel>::finalize()
{
    if(_synchronizer) { _synchronizer->finalize(); }
}

template<class UserModel>
void Model<UserModel>::setInputs()
{
    if(_synchronizer) { _synchronizer->setInputs(); }
}

template<class UserModel>
void Model<UserModel>::getOutputs()
{
    if(_synchronizer) { _synchronizer->getOutputs(); }
}

template<class UserModel>
void Model<UserModel>::simulate()
{
    if(_synchronizer) { _synchronizer->simulate(); }
}

template<class UserModel>
void Model<UserModel>::setSynchronizer(Synchronizer *synchronizer)
{
    _synchronizer = synchronizer;
}

////////////////////////////////////////////////////////////////////////////////
// Auxiliary methods
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
Process* Model<UserModel>::fork(Interface &iface, Message &message)
{
    // IMPORTANT: Newly created process should have a greater PID than the
    // previously launched ones. This guarantees that reset of a control signal
    // (strobe) is done before a new process set the signal.
    Process *process = new Process(_new_pid++, iface, message);
    CPPTESK_ASSERTION(process != 0, "No memory");

    _process_table.insert(std::pair<pid_t, Process*>(process->pid, process));

    return process;
}

template<class UserModel>
void Model<UserModel>::resume(Process *process)
{
    ucontext_t *from;
    ucontext_t *jump;

    CPPTESK_ASSERTION(process != 0, "Parameter process is null");

    process->clk_total++;
    process->clk_count++;

#ifdef CPPTESK_DEBUG_HW
    printf("Model::resume: switch to process %p\n", process);
#endif

    CPPTESK_ASSERTION(!_process_stack.empty(), "Execution flow inconsistency");

    from = _process_stack.top();
    _process_stack.push(process->ucp);
    jump = _process_stack.top();

    swapcontext(from, jump);

    switch(process->status)
    {
    case UNPREDICT:
        { onPreconditionViolation(*process); }
        break;
    case VIOLATION:
        { onPostconditionViolation(*process); }
        break;
    case COMPLETED:
        { onNormalCompletion(*process); }
        break;
    case EXECUTING:
        { onEveryCycle(*process); }
        break;
    default:
        { CPPTESK_ASSERTION(false, "Incorrect process status"); }
    }
}

template<class UserModel>
void Model<UserModel>::suspend(Process *process)
{
    ucontext_t *from;
    ucontext_t *jump;

    CPPTESK_ASSERTION(process != NULL, "Parameter process is null");

#ifdef CPPTESK_DEBUG_HW
    printf("Model::suspend: switch from process %p\n", process);
#endif

    CPPTESK_ASSERTION(_process_stack.size() > 1, "Execution flow inconsistency");

    from = _process_stack.top();
    _process_stack.pop();
    jump = _process_stack.top();

    CPPTESK_ASSERTION(from == process->ucp, "Execution flow inconsistency");

    swapcontext(from, jump);
}

CPPTESK_MODEL(EMPTY_MODEL) {};

}} // namespace cpptesk::hw

