/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <generator/tracer.h>

#include <stdio.h>
#include <string.h>

#include <atl/map.h>
#include <atl/string.h>
#include <c_tracer/c_tracer.h>
#include <ts/engine_tsi.h>

#include <aspectrace.h>
#include <aspect_fsm.h>

/*                       YYYY-MM-DD--HH-MM-SS.NNN */
#define MAX_TIME_STRLEN (5    +3 +2+2+3 +3 +3 +3 +1)

static char* generateTTFileName(const char *basename, const char *cpu_id, const char *extension)
{
    int j;
    char* timemark = (char*)format_time();
    char* ttfilename;

    ttfilename = malloc(strlen(basename) + MAX_TIME_STRLEN + 16);
    assertion(ttfilename != NULL, "generateTTFileName: not enough memory");

    sprintf(ttfilename, "%s_%s_%s.%s", basename, cpu_id, timemark, extension);

    for(j = 1; j < 1000; j++)
    {
        if(!file_exists(ttfilename))
            { break; }

        sprintf(ttfilename, "%s_%s_%s.%03d.%s", basename, cpu_id, timemark, j, extension);
    }

    return ttfilename;
}

bool startTrace_TestEngine()
{
    // UTT trace
    if(!isInTrace())
    {
        startTrace();

        return true;
    }

    return false;
}

void endTrace_TestEngine(bool tracing_was_started_here)
{
    if(tracing_was_started_here)
    {
        // UTT trace
        endTrace();

        // UTT2 trace
        aspectrace_finishTracing();
    }
}

void traceScenarioStart_TestEngine(const char *name, int argc, char **argv)
{
    // UTT trace
    traceScenarioStart(name);
    traceScenarioParameters(argc, (const char**)argv);
    // Deferred Reactions Enabled
    traceTestEngineProperty( "Deferred Reactions Enabled", areDeferredReactionsEnabled() ? "true" : "false" );

    // UTT2 trace
    aspect_fsm_traceFsmStart(name);
}

void traceScenarioEnd_TestEngine(void)
{
    // UTT trace
    traceScenarioEnd();

    // UTT2 trace
    aspect_fsm_traceFsmEnd();
}

bool traceTransitionStart_TestEngine(const char *tran_id, const char *name, Object *value, bool accidental)
{
    bool old = setTraceAccidental(accidental);

    // UTT trace
    traceTransitionStart(tran_id);
    traceScenarioValue("scenario method", "", "", name);

    // UTT2 trace
    aspect_fsm_traceTransitionStart(tran_id, name);
    aspect_fsm_traceIterationVariable("Object*", "value", value != NULL ? r(value) : create_String("NULL"));

    return old;
}

void traceTransitionEnd_TestEngine(const char *tran_id, const char *name, bool accidental)
{
    // UTT trace
    traceTransitionEnd();
    setTraceAccidental(accidental);

    // UTT2 trace
    aspect_fsm_traceTransitionEnd(tran_id, name);
}

void traceState_TestEngine(const char *state_id, const char *state)
{
    // UTT trace
    traceState(state_id);
    traceScenarioValue("state", "", "", state);

    // UTT2 trace
    aspect_fsm_traceState(state_id, state);

#ifdef DEBUG_NETFSM
    printf("[STATE] %s\n", state_id);
#endif
}

void traceProperty_TestEngine(const char *property, const char *value)
{
    // UTT trace
    traceTestEngineProperty(property, value);
}

void traceSystemInfo_TestEngine(const char *name, const char *info)
{
    // UTT trace
    traceSystemInfo(name);
    traceSystemInfo(info);

    // UTT2 trace
    aspect_fsm_traceBrowserMessage(ML_INFO, info);

    // STDOUT
    printf("[SYSTEM] %s: %s\n", name, info);
}

void traceExceptionInfo_TestEngine(const char *name, const char *info)
{
    // UTT trace
    traceExceptionInfo(info);
    traceException(name);

    // UTT2 trace
    aspect_fsm_traceBrowserMessage(ML_ERROR, info);

    // STDOUT
    printf("[EXCEPTION] %s: %s\n", name, info);
}

void traceNondeterminism_TestEngine(const char *start_state_id, int stimulus_id, const char *old_end_state_id, const char *new_end_state_id)
{
    String *msg = format_String("start_state=%s, stimulus_id=%d, old_end_state=%s, new_end_state=%s",
        start_state_id, stimulus_id, old_end_state_id, new_end_state_id);

    traceSystemInfo_TestEngine("NONDETERMINISM_FOUND", toCharArray_String(r(msg)));
    destroy(msg);
}

void traceTestSummary_TestEngine(Map *summary)
{
    int i;

    for(i = 0; i < size_Map(r(summary)); i++)
    {
        String *key = key_Map(r(summary), i);
        String *val = get_Map(r(summary), r(key));
        String *msg = format_String("%s: %s", toCharArray_String(r(key)), toCharArray_String(r(val)));

        // UTT trace
        traceSystemInfo(toCharArray_String(r(msg)));

        // STDOUT
        printf("%s\n", toCharArray_String(r(msg)));
        destroy(msg);
    }

    // UTT2 trace
    aspect_fsm_traceSummary(r(summary));
}

void flushTrace_TestEngine(void)
{
    // UTT trace
    flushTrace();

    // STDOUT
    fflush(stdout);
}

void processTraceParameters_TestEngine(const char *name, const char *cpu_id, int argc, char **argv)
{
    int i;

    bool nt_flag = false;
    bool tf_flag = false;
    bool tt_flag = false;
    bool ts_flag = false;

    assertion(name != NULL,              "processAspectraceParameters: name == NULL");
    assertion(argc >= 0,                 "processAspectraceParameters: argc < 0");
    assertion(argv != NULL || argc == 0, "processAspectraceParameters: argv == NULL && argc > 0");

    for(i = 0; i < argc; i++)
    {
        if(!strcmp("-nt2", argv[i]))
        {
            nt_flag = true;

            assertion(!tf_flag,   "Bad aspectrace parameters: -tf2 and -nt2 options are mutually exclusive");
            assertion(!tt_flag,   "Bad aspectrace parameters: -tt2 and -nt2 options are mutually exclusive");
            assertion(!ts_flag,   "Bad aspectrace parameters: -ts2 and -nt2 options are mutually exclusive");
        }
        else if(!strcmp("-tf2", argv[i]))
        {
            tf_flag = true;

            assertion(!nt_flag,   "Bad aspectrace parameters: -tf2 and -nt2 options are mutually exclusive");
            assertion(++i < argc, "Bad aspectrace parameters: after -tf2 should be a file name");

            aspectrace_addXmlTrace(argv[i]);
        }
        else if(!strcmp("-tt2", argv[i]) && !tt_flag)
        {
            tt_flag = true;

            assertion(!nt_flag,   "Bad aspectrace parameters: -tt2 and -nt2 options are mutually exclusive");

            aspectrace_addXmlTrace(generateTTFileName(name, cpu_id, "utt2"));
        }
        else if(!strcmp("-ts2", argv[i]))
        {
            char *p = NULL;

            ts_flag = true;

            assertion(!nt_flag,   "Bad aspectrace parameters: -ts2 and -nt2 options are mutually exclusive");
            assertion(++i < argc, "Bad aspectrace parameters: after -ts2 should be a TCP/IP address (<host>:<port>)");

            p = strchr(argv[i], ':');
            assertion(p != NULL,  "Bad aspectrace parameters: after -ts2 should be a TCP/IP address (<host>:<port>)");

            *p = '\0';
            aspectrace_addTcpSocket(argv[i], (unsigned short)atoi(p + 1));
            *p = ':';
        }
    }
 
    if(!nt_flag && !tf_flag && !tt_flag && !ts_flag)
        { aspectrace_addXmlTrace(generateTTFileName(name, cpu_id, "utt2")); }
}
